"""
Custom exceptions for easier error handling

General exceptions can be used for convenient error catching
example:

class GeneralException(Exception):
    ...

class SpecificException(GeneralException):
    ...

try:
    raise SpecificException
except GeneralException:
    ...
"""


class CustomException(Exception):
    """ User defined exception"""

    def __init__(self, *args):
        self.text = "An undefined custom exception has been raised"

    def __str__(self):
        """ raising an exception invokes the __str__ override """
        return self.text


class FileHandlerException(CustomException):
    """ Exceptions for reading files """

    def __init__(self, *args):
        self.text = "Exception occurred during file read"


# ----------------- File Handling ----------------
class UnsupportedFileExtension(FileHandlerException):
    """ File extension handler doesn't exist """

    def __init__(self, bad_file: str):
        self.text = f"Unable to load file: \"{bad_file}\", unsupported extension"


class FileNotFoundError(FileHandlerException, FileNotFoundError):
    """ Override FileNotFound as a custom exception """
    def __init__(self, bad_file: str):
        self.text = f"Unable to load file: \"{bad_file}\", file does not exist"



"""
Entrypoint to the discovery project
Reads a filesystem and tries to make some analysis based off of it
"""
import itertools
import json
import os

import yaml
import logging.config
from typing import Type

# set up local logging before importing local libs
# TODO: do this a better way
if __name__ == "__main__":
    with open('logging_conf.yaml', 'r') as f:
        config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)

    logger = logging.getLogger(__name__)

from discovery import metadata
from discovery.utils.visualizer import Visualizer
from discovery.data_matching.data_match_interface import DataMatcher
from discovery.data_matching.dataframe_matcher import DataFrameMatcher
from discovery.data_matching.matching_methods import *


class DiscoveryClient:
    def __init__(self, discovery_config: dict):
        self.config = discovery_config
        self.loaded_catalogue: dict[str:metadata.CatalogueItem] = {}
        self.visualiser = Visualizer()
        self.dataframe_matcher = DataFrameMatcher()

    def create_visual(self, pathname: str):
        """Build a visual based on stored metadata
        Args:
            pathname: The path to draw the visual

        Returns: None
        """
        # TODO: find a fancy python way to do this
        metadata_list = []
        for item in self.loaded_catalogue.values():
            metadata_list.append(item.get_metadata())
        self.visualiser.draw(metadata_list, pathname)

    def construct_relationships(self, methods: list[Type[DataMatcher]]):
        """Construct a set of relationships between columns in the dataframes
        Args:
            methods: A list of methods to use when constructing the relationships

        Returns: None
        """
        for item1, item2 in itertools.combinations(self.loaded_catalogue.values(), 2):
            self._match_metadata(item1, item2, methods)

    def _match_metadata(self, origin_item: metadata.CatalogueItem, target_item: metadata.CatalogueItem, methods):
        """Run the dataframe matcher with the two given dataframes
        Update the metadata to reflect the changes
        """
        # TODO: paged matching for large dataframes
        origin_data = origin_item.get_data()
        target_data = target_item.get_data()
        origin_metadata = origin_item.get_metadata()
        target_metadata = target_item.get_metadata()
        for col1, col2 in itertools.product(origin_metadata.columns.values(), target_metadata.columns.values()):
            # results = self.dataframe_matcher.match_dataframes(reference_dataframe, subject_dataframe)
            result = self.dataframe_matcher.match_columns(
                methods=methods,
                col_meta1=col1,
                col_meta2=col2,
                series1=origin_data[col1.name],
                series2=target_data[col2.name]
            )
            # keep just the average for now
            result_average = result[1]
            col1.add_relationship(
                result_average, target_item.get_checksum(), col2.name
            )

    def add_catalogue_item(self, catalogue_item: metadata.CatalogueItem):
        """Adds a metadata item to the ones loaded in memory

        Args:
            catalogue_item: The desired catalogue item to add

        Returns: None
        """
        self.loaded_catalogue[catalogue_item.get_id()] = catalogue_item

    def write_catalogue(self, path: str):
        """Write all loaded metadata to a json file

        WARNING: overwrites previous entry

        Args:
            path: The path to write metadata to

        Returns: None
        """
        catalogue_items = [
            item.write_metadata()
            for item in self.loaded_catalogue.values()
        ]
        with open(path, 'w') as catalogue_file:
            catalogue_file.write(json.dumps(catalogue_items))

    def load_file(self, path, catalogue_metadata=None, construct_metadata=False):
        """Uses the from_csv method to load a csv file into the catalogue

        Args:
            path: The path to the file

        Returns: None
        """
        new_item = metadata.from_csv(path)
        new_item.rebuild_metadata_object()
        self.add_catalogue_item(new_item)

    def load_metadata(self, path):
        """Uses the from_metdata method to load in previously created metadata items
        as catalogue items

        Args:
            path: The path to the metadata file

        Returns: None
        """
        new_items = metadata.from_metadata_json(path)
        for new_item in new_items:
            new_item.rebuild_metadata_object()
            self.add_catalogue_item(new_item)

    def scan_local_filesystem(self, path):
        """Read in data from a local filesystem, given a path

        Args:
            path: The root path of the data

        Returns: A generator of each item produced
        """
        for root, dirs, files in os.walk(path):
            for filename in files:
                if filename.endswith('csv'):
                    new_item = metadata.from_csv(os.path.join(root, filename))
                    new_item.rebuild_metadata_object()
                    yield new_item
                    self.add_catalogue_item(new_item)


if __name__ == "__main__":
    # locally test the mock filesystem
    launch_config = yaml.safe_load(open("launch_config.yaml"))
    discovery_instance = DiscoveryClient(launch_config)

    from utils.datagen import FakeDataGen

    fake_data = FakeDataGen()
    # fake_files = fake_data.build_df_to_file(1000, "matcher_test", index_type="categoric", continuous_data=5,
    #                                         file_spread=2)
    # discovery_instance.load_file(fake_files[0])
    # discovery_instance.load_file(fake_files[1])
    discovery_instance.load_metadata("newtest.json")

    discovery_instance.construct_relationships(
        methods=[
            MatchColumnNamesLCS,
            MatchIdenticalColumnNames
        ]
    )

    discovery_instance.write_catalogue("newtest.json")

    discovery_instance.create_visual("test_visual")



"""
Calculate the normalized distance measure between two numerical columns using Dynamic Time Warping
"""

from dtaidistance import dtw
import pandas as pd
from discovery.data_matching.data_match_interface import DataMatcher


class MatchDataDynamicTimeWarping(DataMatcher):
    @staticmethod
    def run_process(series1: pd.Series = None, series2: pd.Series = None, **kwargs):
        max_distance = len(series1) * max(series1)

        if pd.api.types.is_numeric_dtype(series1) and pd.api.types.is_numeric_dtype(series2):
            return (max_distance - dtw.distance(series1, series2)) / max_distance * 100
        return 0



"""
Implements standard methods expected of data matching techniques

NOTE:
    we're filling out unused parameters on purpose
TODO:
    add decorators for enforcing int values
"""
import pandas as pd
from discovery.metadata import ColMetadata, CatalogueMetadata
from discovery import metadata


class DataMatcher:
    def __init__(self):
        self.name = self.__class__.__qualname__

    def match_columns(
            self,
            col_meta1: metadata.ColMetadata,
            col_meta2: metadata.ColMetadata,
            series1: pd.Series = None,
            series2: pd.Series = None,
            metadata1: metadata.CatalogueItem = None,
            metadata2: metadata.CatalogueItem = None

    ) -> int:
        incoming = locals()
        incoming.pop('self')
        return self.run_process(**incoming)

    @staticmethod
    def run_process(
            col_meta1: ColMetadata,
            col_meta2: ColMetadata,
            series1: pd.Series = None,
            series2: pd.Series = None,
            metadata1: CatalogueMetadata = None,
            metadata2: CatalogueMetadata = None
    ) -> int:
        """
        Run the actual implemented process
        Reiterate the match_columns parameters here, as IDE's can pick it up
        """
        raise NotImplementedError



"""
A set of functions that will build catalogue items for us
Unfortunately it's hard to find a clean way to systematically attach json to data builders, our approach is to have
non backwards-compatible readers that are versioned for the current implementation of this file
"""
import copy
import json
import logging
from jsonschema import validate

from discovery.metadata.catalogue_item import CatalogueItem
from discovery.metadata.catalogue_metadata import CatalogueMetadata
from discovery.utils.data_handling.local_csv_handler import LocalCSVHandler
from discovery.utils.data_handling import data_size
from discovery.utils.data_handling.catalogue_data import *

logger = logging.getLogger(__name__)

# Ensure that versions remain consistent
BUILDER_VERSION = 1


def from_csv(
        path: str,
        dataframe_metadata: CatalogueMetadata = None
):
    """Construct a metadata object by providing a csv file"""
    data = LocalCSVHandler(path)
    if dataframe_metadata is None:
        # if no metadata is provided, build an empty one
        dataframe_metadata = CatalogueMetadata(data_manifest=data.get_manifest())
    return CatalogueItem(metadata=dataframe_metadata, data=data)


DICT_LOADERS = {
    "LocalCSVHandler": LocalCSVHandler,
    "default": CatalogueData
}
LOADER_DICTS = {val: key for key, val in DICT_LOADERS.items()}
DICT_DATA_SIZES = {
    "file": data_size.FileDataItemSize,
    "default": data_size.DataItemSize
}


def from_metadata_json(
        metadata_json_path: str
) -> [CatalogueItem]:
    """Construct a set of catalogue items from a json descriptor file
    For now, we can just re-run the loader
    """
    metadata_json = json.load(open(metadata_json_path))
    new_items = []

    for catalogue_pair in metadata_json:
        item_id, catalogue_json = list(catalogue_pair.items())[0]
        data_handling = catalogue_json['data_handling']

        new_metadata = CatalogueMetadata(
            item_id=item_id,
            data_manifest=copy.copy(data_handling),
            tags=catalogue_json['tags']
        )

        data_size_attributes = data_handling.pop('data_size')
        data_size_type = data_size_attributes.pop('size_representation', '')
        data_size_object = DICT_DATA_SIZES.get(data_size_type, data_size.DataItemSize)(**data_size_attributes)
        data_handler = DICT_LOADERS[data_handling.pop('loader', 'default')]
        new_data = data_handler(
            data_size=data_size_object,
            **data_handling
        )

        new_items.append(CatalogueItem(metadata=new_metadata, data=new_data))

    return new_items



import uuid
from typing import Union
from abc import ABC


class Relationship:
    certainty: int
    target_hash: int
    target_column_name: str

    def __init__(self, certainty, target_file_hash, target_column_name):
        self.certainty = certainty
        self.target_hash = target_file_hash
        self.target_column_name = target_column_name


class ColMetadata(ABC):
    name: str
    col_type: str
    relationships: list[Relationship]

    def __init__(self, name: str, col_type: str, continuity: float):
        self.name = name
        self.col_type = col_type
        self.continuity = continuity
        self.relationships = []

    def get_attributes(self, stringify: bool):
        """Get the attributes of the column as a dictionary

        Returns: A dictionary of the column attributes
        """
        col_type = self.col_type if not stringify else str(self.col_type)
        return {
            "type": col_type,
            "continuity": self.continuity,
        }

    def add_relationship(self, certainty, target_id, target_column_name):
        self.relationships.append(Relationship(certainty, target_id, target_column_name))


class NumericColMetadata(ColMetadata):
    mean: Union[float, None]
    minimum: any
    maximum: any

    def __init__(self, name: str, col_type: str, continuity: float,
                 mean: Union[int, float, None], min_val, max_val):
        ColMetadata.__init__(self, name, col_type, continuity)
        self.mean = mean
        self.minimum = min_val
        self.maximum = max_val

    def get_attributes(self, stringify: bool):
        """Return the mean/min/max of the column for numeric columns

        Returns: A dictionary of the column attributes
        """
        col_type = self.col_type if not stringify else str(self.col_type)
        return {
            "type": col_type,
            "continuity": self.continuity,
            "mean": self.mean,
            "max": self.maximum,
            "min": self.minimum,
        }


class CategoricalColMetadata(ColMetadata):
    def __init__(self, name: str, col_type: str, continuity: float, columns=None):
        ColMetadata.__init__(self, name, col_type, continuity)


class CatalogueMetadata:
    item_id: uuid.UUID
    data_manifest: dict
    columns: dict[str:ColMetadata] = []
    tags: dict = {}

    def __init__(
            self,
            data_manifest: dict,
            item_id: uuid.UUID = None,
            columns: dict[str, ColMetadata] | None = None,
            tags: dict | None = None
    ):
        self.item_id = item_id
        self.data_manifest = data_manifest
        self.columns = columns or {}
        self.tags = tags or {}



from discovery.data_matching.matching_methods.match_column_name_lcs import MatchColumnNamesLCS
from discovery.data_matching.matching_methods.match_column_name_levenshtein import MatchColumnNamesLevenshtein
from discovery.data_matching.matching_methods.match_column_name_wordnet import MatchColumnNamesWordnet
from discovery.data_matching.matching_methods.match_data_dynamic_time_warping import MatchDataDynamicTimeWarping
from discovery.data_matching.matching_methods.match_data_pearson_coefficient import MatchDataPearsonCoefficient
from discovery.data_matching.matching_methods.match_data_two_sample_t_test import MatchDataTwoSampleTTest
from discovery.data_matching.matching_methods.match_identical_rows import MatchIdenticalRows
from discovery.data_matching.matching_methods.match_identical_column_names import MatchIdenticalColumnNames


"""
Calculate the similarity of 2 column names using WordNet
"""

from nltk.corpus import wordnet
from itertools import product

from discovery.data_matching.data_match_interface import DataMatcher


class MatchColumnNamesWordnet(DataMatcher):
    @staticmethod
    def run_process(col_meta1, col_meta2, **kwargs):
        synset_first = wordnet.synsets(col_meta1)
        synset_second = wordnet.synsets(col_meta2)
        wordnet_ratio = 0
        if len(synset_first) > 0 and len(synset_second) > 0:
            wordnet_ratio = max(
                wordnet.wup_similarity(s1, s2) for s1, s2 in product(synset_first, synset_second)) * 100

        return wordnet_ratio



from discovery.metadata.catalogue_item import CatalogueItem
from discovery.metadata.catalogue_metadata import CatalogueMetadata
from discovery.metadata.catalogue_metadata import ColMetadata
from discovery.metadata.catalogue_metadata import CategoricalColMetadata
from discovery.metadata.catalogue_metadata import NumericColMetadata

from discovery.metadata.catalogue_item_builders import *


"""
When applied to a function, this will raise an exception if the input types don't match the type hints
The type enforcer is expensive to run (~1ms per run), so use it only for user facing functions
"""

import inspect


def type_enforcer(input_function):
    """
        raises a type error if the given arguments don't match the signature
        try to do as much at initialisation, to lower performance cost (and increase memory cost)
    """
    signature = inspect.signature(input_function)
    required_args = [signature.parameters[x].annotation for x in signature.parameters.keys()]
    required_kwargs = {x: signature.parameters[x].annotation for x in signature.parameters.keys() if
                       signature.parameters[x].default != signature.empty}

    def function_wrapper(*args, **kwargs):
        # test positional arguments
        correct_args = [isinstance(x, required_args[n]) for n, x in enumerate(args) if
                        required_args[n] is not inspect._empty]
        if not all(correct_args):
            raise TypeError

        # test keyword arguments
        correct_kwargs = [isinstance(val, required_kwargs[key]) for key, val in kwargs.items()]
        if not all(correct_kwargs):
            raise TypeError

        return input_function(*args, **kwargs)

    return function_wrapper


if __name__ == "__main__":
    # local imports for tests
    from time import perf_counter_ns

    ns_to_ms = 1e-6


    # local tests to check type enforcement and give examples of how to use it
    @type_enforcer
    def simple_test(x: int, y: str, a: int = 2, b: str = "x"):
        return x, y, a, b


    print('\n' + '=' * 20)
    print("Running simple tests")
    print("Running passing simple test")
    simple_test(1, "string", 1, b="string")
    print("Passing state successfully passed\n")
    print("Running failing simple test")
    try:
        # your IDE should highlight the following statement
        simple_test("string", 1, a='x')
    except TypeError as e:
        print("Failing state successfully failed\n")


    @type_enforcer
    def multi_types(x: int | str, y, z: dict | tuple = ()):
        return x, y, z


    print('\n' + '=' * 20)
    print("Running multiple type tests")
    print("Running passing multiple type test")
    multi_types("string", 1.0, {'key': 'value'})
    print("Passing state successfully passed\n")
    print("Running failing multiple type test")
    try:
        multi_types(5, ("tuple"), [1])
    except TypeError as e:
        print("Failing state successfully failed\n")


    # benchmark tests
    def benchmark_control(x: int):
        return x + 1


    @type_enforcer
    def benchmark_test(x: int):
        return x + 1


    print('\n' + '=' * 20)
    print("Running benchmark tests")

    control_start_time = perf_counter_ns()
    for _ in range(100000):
        benchmark_control(1)
    control_end_time = perf_counter_ns() - control_start_time

    print(f"Control benchmark complete, elapsed time: {control_end_time * ns_to_ms}s")

    test_start_time = perf_counter_ns()
    for _ in range(100000):
        benchmark_test(1)
    test_end_time = perf_counter_ns() - test_start_time

    print(f"Simple benchmark complete, elapsed time: {test_end_time * ns_to_ms}s")



"""
The simplest possible solution to matching column data
"""

from discovery.data_matching.data_match_interface import DataMatcher


class MatchIdenticalColumnNames(DataMatcher):
    @staticmethod
    def run_process(col_meta1, col_meta2, **kwargs):
        return int(col_meta1.name == col_meta2.name) * 100



import uuid
from typing import Iterable

import pandas as pd
import pandas.errors

from discovery.utils.data_handling.catalogue_data import CatalogueData
from discovery.utils.data_handling.data_size import *


class LocalCSVHandler(CatalogueData):
    def __init__(self, path, checksum: uuid.UUID = None, data_size: DataItemSize = None):
        # The specification to be used to replicate reading
        self.data_size = data_size
        self.checksum = checksum
        self.path = path

    def get_manifest(self, update=True) -> dict:
        """Return a json representation of the local CSV that can be used to pull the data again
        We use the checksum as part of this representation, as it can allow us to build a resiliency to path changes

        Note: we do not store any data on the filesystem, this allows replication across file systems
        """
        return {
            "loader": self.__class__.__name__,
            "checksum": self.get_checksum(),
            "path": self.path,
            "data_size": self.get_data_size(update=update, json_representation=True)
        }

    def get_data_size(self, update=False, json_representation=False) -> DataItemSize | dict:
        """Return a size based on the amount of rows and bytes in the data"""
        if update:
            dataframe = next(self.get_data())
            rows = dataframe.shape[0]
            byte_num = int(dataframe.memory_usage(index=True).sum())
            self.data_size = FileDataItemSize(
                no_of_rows=rows,
                no_of_bytes=byte_num
            )

        if json_representation:
            return self.data_size.get_attributes()
        return self.data_size

    def get_data(self, rows: int | None = None) -> pd.DataFrame | Iterable[pd.DataFrame]:
        """Return a generator for the data

        Use lower row counts to preserve memory, and higher row counts for more efficient reading
        Args:
            rows: The amount of rows to generate on each iteration

        Returns: A generator of the dataframe, returning the amount of rows specified each time

        """
        if rows is None:
            yield pd.read_csv(self.path)

        seek_row = 0
        header_df = pd.read_csv(self.path, skiprows=seek_row, nrows=rows)
        df_columns = header_df.columns
        yield header_df

        while True:
            seek_row += rows
            try:
                yield pd.read_csv(self.path, skiprows=seek_row, nrows=rows, usecols=df_columns)
            except pandas.errors.EmptyDataError:
                break



import uuid
from typing import Iterable

import pandas as pd
from pandas.api.types import is_numeric_dtype

from discovery.metadata.catalogue_metadata import CatalogueMetadata, ColMetadata, NumericColMetadata, \
    CategoricalColMetadata
from discovery.utils.data_handling.catalogue_data import CatalogueData


class CatalogueItem:
    """Catalogue item abstraction used to organise data into more understandable sets"""

    def __init__(
            self,
            metadata: CatalogueMetadata,
            data: CatalogueData
    ):
        self._metadata = metadata
        self._data = data
        self._id = uuid.uuid4()

    def get_id(self) -> uuid:
        """return the UUID ID of this catalogue item

        Returns:
            The UUID of the class as a UUID object
        """
        return self._id

    def update_tags(self, tag_update: dict):
        """Updates the tags that have been assigned to this catalogue item

        WARNING: overwrites previous values

        Args:
            tag_update (dict): A dictionary of {tag_name: tag_value}

        Returns:
            None
        """
        self._metadata.tags.update(tag_update)

    def get_data(self, rows: int | None = None) -> pd.DataFrame | Iterable[pd.DataFrame]:
        """Return a generator for the data

        Use lower row counts to preserve memory, and higher row counts for more efficient reading
        If the row amount is not specified, return the entire dataframe
        Args:
            rows: The amount of rows to generate on each iteration

        Returns: A generator of the dataframe, returning the amount of rows specified each time

        """
        data_generator = self._data.get_data(rows)
        if rows is None:
            return next(data_generator)
        return data_generator

    def get_checksum(self, update=True):
        """Return a checksum for the given dataframe

        Allows for abstracting checksum calculations into the reader itself
        If update is set to True, the checksum needs to be rebuilt,
        not rebuilding the checksum can be used for static data
        """
        return self._data.get_checksum(update=update)


    def get_metadata(self) -> CatalogueMetadata:
        """Return the metadata for this catalogue item

        Returns: A CatalogueMetadata object
        """
        return self._metadata

    def write_metadata(self) -> dict:
        """Return the metadata as a json compatible object

        Returns: A dictionary representation of the metadata
        """
        return {
            f"{str(self.get_id())}": {
                "data_handling": self._data.get_manifest(),
                "tags": {
                    name: value
                    for name, value in self._metadata.tags.items()
                },
                "columns": {
                    column_name: _write_metadata_column(column)
                    for column_name, column in self._metadata.columns.items()
                }
            }
        }

    def rebuild_metadata_object(self):
        """This method rebuilds all metadata that is reliant on the dataframe
        TODO: currently unable to read dataframe as chunks

        Metadata being regenerated is as follows:
        - data_checksum
        - data size
        - number of rows
        - column metadata

        This method can be used for when data gets updated, or the checksum does not match the loaded data
        Returns: None
        """
        dataframe: pd.DataFrame = self.get_data()
        column_metadata = {}
        for column in dataframe.columns:
            column_metadata[column] = resolve_metadata_column(column, dataframe[column])

        self._metadata = CatalogueMetadata(
            item_id=self.get_id(),
            data_manifest=self._data.get_manifest(),
            columns=column_metadata,
            tags=self._metadata.tags
        )


def resolve_metadata_column(name: str, series: pd.Series) -> ColMetadata:
    """Creates column metadata based on a column name, and column values
    Args:
        name: The name of the column
        series: The series to be evaluated

    Returns: A Col Metadata object
    """
    continuity = series.nunique() / series.count()
    if is_numeric_dtype(series):
        return NumericColMetadata(name, series.dtype, continuity, *get_series_statistical_values(series))
    return CategoricalColMetadata(name, series.dtype, continuity)


def _write_metadata_column(column: ColMetadata):
    """Return a column as a json compatible object"""
    return {
        "name": column.name,
        "attributes": column.get_attributes(stringify=True),
        "relationships": [{
            "certainty": relationship.certainty,
            "target_hash": relationship.target_hash,
            "target_column_name": relationship.target_column_name
        } for relationship in column.relationships]
    }


def get_series_statistical_values(series: pd.Series):
    """If the column is numeric, get statistics on it

    Args:
        series: A numeric pandas series

    Returns: A tuple of the following:
                The average (mean) value of the column
                The minimum value of the column
                The maximum value of the column
                The probability that the column represents continuous data

    """
    return (
        series.mean(),
        series.min(),
        series.max()
    )



"""
Takes two dataframes and tries to match them column by column
"""
import logging
from typing import Type

# TODO: currently not working out of the box, must be run with the following statements...
# >>> import nltk
# >>> nltk.download('wordnet')
# >>> nltk.download('omw-1.4')
import pandas as pd
import numpy

from discovery.data_matching.data_match_interface import DataMatcher
from discovery.metadata import CatalogueMetadata, NumericColMetadata, ColMetadata

logger = logging.getLogger(__name__)


class DataFrameMatcher:
    @staticmethod
    def match_columns(
            methods: list[Type[DataMatcher]],
            col_meta1: ColMetadata,
            col_meta2: ColMetadata,
            series1: pd.Series = None,
            series2: pd.Series = None,
            metadata1: CatalogueMetadata = None,
            metadata2: CatalogueMetadata = None,
            weights: list[int] = None
    ) -> (dict, int):
        """
        Calculate the similarity of 2 columns using a combination of methods
        Weights is given as an ordered list of weights to apply to each method, respective to the order of the methods

        Returns a dictionary of individual similarity values, followed by an overall average
        """

        weights = weights or []
        # forward fill weights with 1's
        filled_weights = [1] * len(methods)
        filled_weights[:len(weights)] = weights
        similarity_ref = {}
        overall_similarity = 0

        for index, method in enumerate(methods):
            try:
                method_instance = method()

                logger.debug(f"Matching columns using {method_instance.name}")
                similarity = method_instance.run_process(**{
                    "col_meta1": col_meta1,
                    "col_meta2": col_meta2,
                    "series1": series1,
                    "series2": series2,
                    "metadata1": metadata1,
                    "metadata2": metadata2
                })
                # update the similarity reference with applied method and weight
                similarity_ref.update({method_instance.name: (similarity, filled_weights[index])})
                overall_similarity += similarity * filled_weights[index]

            except Exception as exc:
                logger.error(f"Something unexpected went wrong running method {method.__qualname__}, skipping...")
                logger.error(exc)
                # something went wrong, invalidate the method and ensure the weight is correct
                filled_weights[index] = 0

        # get the average similarity, if overall_similarity is zero then something probably went wrong, set to zero
        average_similarity = (overall_similarity / sum(filled_weights)) if overall_similarity else 0
        logger.debug("Matching complete")

        return similarity_ref, average_similarity

    @staticmethod
    def match_column_in_dataframe(df1: pd.DataFrame, column1: NumericColMetadata,
                                  metadata2: CatalogueMetadata, df2: pd.DataFrame):
        """
        Find the best match of a column in another dataframe
        :param df1:
        :param column1:
        :param metadata2:
        :param df2:
        :return:
        """
        scores = {}
        no_of_tests = {}
        average_differences = {}
        minimum_differences = {}
        maximum_differences = {}
        for column2 in metadata2.columns:
            # LCS name test
            lcs_percentage = DataFrameMatcher.match_name_lcs(column1.name, column2.name)

            # Levenshtein name test
            levenshtein_percentage = DataFrameMatcher.match_name_levenshtein(column1.name, column2.name)

            # Data type test
            data_type_matches = 100 if column1.col_type == column2.col_type else 0

            # Continuity test
            continuity_percentage = 100 * (1 - abs(column1.continuity - column2.continuity))

            # Numerical values test
            numerical_percentage = 100 * (1 - abs(column1.is_numeric_percentage -
                                                  column2.is_numeric_percentage))

            # Average test
            if column1.mean is not None and column2.mean is not None:
                average_differences[column2.name] = abs(column1.mean - column2.mean)

            # Min test
            try:
                minimum1 = float(column1.minimum)
                minimum2 = float(column2.minimum)
                minimum_differences[column2.name] = abs(minimum1 - minimum2)
            except ValueError:
                pass

            # Max test
            try:
                maximum1 = float(column1.maximum)
                maximum2 = float(column2.maximum)
                maximum_differences[column2.name] = abs(maximum1 - maximum2)
            except ValueError:
                pass

            # Average similarity
            average_similarity = lcs_percentage + levenshtein_percentage + data_type_matches + \
                                 continuity_percentage + numerical_percentage

            scores[column2.name] = average_similarity
            no_of_tests[column2.name] = 5

        # Normalize the average differences
        normalized_average_differences = DataFrameMatcher.normalize_values(average_differences)
        for name, similarity in normalized_average_differences.items():
            scores[name] += similarity
            no_of_tests[name] += 1

        # Normalize the minimum differences
        normalized_minimum_differences = DataFrameMatcher.normalize_values(minimum_differences)
        for name, similarity in normalized_minimum_differences.items():
            scores[name] += similarity
            no_of_tests[name] += 1

        # Normalize the maximum differences
        normalized_maximum_differences = DataFrameMatcher.normalize_values(maximum_differences)
        for name, similarity in normalized_maximum_differences.items():
            scores[name] += similarity
            no_of_tests[name] += 1

        for column2 in metadata2.columns:
            scores[column2.name] /= no_of_tests[column2.name]

        best_similarity = 0
        best_name = ''
        for name, similarity in scores.items():
            if similarity > best_similarity:
                best_similarity = similarity
                best_name = name

        return best_name, best_similarity

    @staticmethod
    def normalize_values(dictionary: dict):
        normalized_dictionary = {}
        min_value = numpy.inf
        max_value = 0
        for index, value in dictionary.items():
            if value < min_value:
                min_value = value
            if value > max_value:
                max_value = value

        for index, value in dictionary.items():
            normalized_dictionary[index] = (1 - (value - min_value) / (max_value - min_value)) * 100

        return normalized_dictionary



"""
Determine if the unknown population means of the two groups are equal
"""

import numpy as np
import scipy.stats as stats

from discovery.data_matching.data_match_interface import DataMatcher


class MatchDataTwoSampleTTest(DataMatcher):
    @staticmethod
    def run_process(series1, series2, **kwargs):
        # Normalize the data
        series1 = series1 / np.max(np.abs(series1), axis=0)
        series2 = series2 / np.max(np.abs(series2), axis=0)

        # First, we determine if the 2 groups have the same variance
        # A ratio of less than 4:1 indicates we should consider the variances equal
        variance_a = np.var(series1)
        variance_b = np.var(series2)
        if min(variance_a, variance_b) == 0:
            equal_variances = max(variance_a, variance_b) < 4
        else:
            equal_variances = (max(variance_a, variance_b) / min(variance_a, variance_b)) < 4

        # Perform the two sample T-test
        result_statistic, result_p_value = stats.ttest_ind(series1, series2, equal_var=equal_variances)

        # If the p-value is greater than 0.05, we accept the null hypothesis that the mean of the two groups is equal
        # Otherwise, we reject it, and claim that the means are different
        return result_p_value > 0.05



"""
Calculate the Pearson correlation coefficient between the 2 columns if they are both numerical
"""


from discovery.data_matching.data_match_interface import DataMatcher


class MatchDataPearsonCoefficient(DataMatcher):
    @staticmethod
    def run_process(series1, series2, **kwargs):
        pearson_correlation_coefficient = abs(series1.corr(series2)) * 100
        return pearson_correlation_coefficient



"""
Calculate a similarity percentage with set operations (intersection vs union) for (possible) categorical data
"""

import numpy
from discovery.data_matching.data_match_interface import DataMatcher


class MatchIdenticalRows(DataMatcher):
    @staticmethod
    def run_process(series1, series2, **kwargs):
        stringified_column_a = series1.apply(str)
        stringified_column_b = series2.apply(str)

        categorical_similarity = len(numpy.intersect1d(stringified_column_a, stringified_column_b)) / len(
            numpy.union1d(stringified_column_a, stringified_column_b)) * 100
        return categorical_similarity



"""
Calculate the similarity of 2 column names using the Longest Contiguous Matching Subsequence
"""

from difflib import SequenceMatcher

from discovery.data_matching.data_match_interface import DataMatcher


class MatchColumnNamesLCS(DataMatcher):
    @staticmethod
    def run_process(col_meta1, col_meta2, **kwargs):
        return SequenceMatcher(None, col_meta1.name, col_meta2.name).ratio() * 100



"""
Contains representations for data size
"""


class DataItemSize:
    attributes = {
        "size_representation": "undefined",
    }

    def __init__(self, **kwargs):
        pass

    def get_attributes(self):
        return self.attributes


class FileDataItemSize(DataItemSize):
    def __init__(self, no_of_rows, no_of_bytes):
        self.attributes = {
            "size_representation": "file",
            "no_of_rows": no_of_rows,
            "no_of_bytes": no_of_bytes
        }



"""
Calculate the similarity of 2 column names using the Levenshtein distance
"""

from Levenshtein import ratio

from discovery.data_matching.data_match_interface import DataMatcher


class MatchColumnNamesLevenshtein(DataMatcher):
    @staticmethod
    def run_process(col_meta1, col_meta2, **kwargs):
        return ratio(col_meta1.name, col_meta2.name) * 100



from typing import Generator, Callable

import pandas as pd
from pandas.util import hash_pandas_object

from .data_size import DataItemSize


class CatalogueData:
    """Catalogue data abstraction which allows interaction with the data, ideally avoids storing data in memory"""
    checksum = None
    data_size: DataItemSize = DataItemSize()

    def get_manifest(self, update=True) -> dict:
        """Return a dictionary representation of the data that is being used"""
        return {
            "data_loader": "undefined",
            "data_size": self.get_data_size(update=update, json_representation=True)
        }

    def get_data_size(self, update=True, json_representation=False) -> DataItemSize | dict:
        """Return a representation of the data size, this is usually a row count and byte count,
        but for more complex data implementations it might have to be something more abstract

        if update is set to False, just use the one already made, but only if one has already been generated
        if json_representation is set to True, return a dictionary representation of the size instead
        """
        if update and self.data_size:
            self.data_size = DataItemSize()

        if json_representation:
            return self.data_size.get_attributes()
        return self.data_size

    def get_data(self, rows: int = None) -> Generator:
        raise NotImplemented

    def get_checksum(self, update=True) -> int:
        """Return a checksum for the given dataframe

        Allows for abstracting checksum calculations into the reader itself
        If update is set to True, the checksum needs to be rebuilt,
        not rebuilding the checksum can be used for static data
        """
        if update:
            dataframe = next(self.get_data())
            self.checksum = int(hash_pandas_object(dataframe).sum())
        return self.checksum



import math
import random
import time
import uuid
import faker
import numpy as np
import pandas as pd
from typing import Callable, Any
from uuid import UUID


class FakeDataGen:
    def __init__(self):
        self.fake = faker.Faker()
        self.index_types = {
            'datetime': self._generate_datetime_index,
            'counter': self._generate_counter_index,
            'categoric': self._generate_unique_categoric_index
        }

    def build_df_to_file(self, rows, path, index_type='datetime', continuous_data=1, categoric_data=1, file_spread=1,
                         file_extension='csv'):
        """
            Create the dataframe as a file or set of files
        """
        generated_df = self.build_df(rows=rows, index_type=index_type, continuous_data=continuous_data,
                                     categoric_data=categoric_data)
        step = rows // file_spread
        generated_files = []
        for file_number in range(file_spread):
            filename = f"{path}_{file_number}.csv"
            generated_df[step * file_number:step * (file_number + 1)].to_csv(filename)
            generated_files.append(filename)

        return generated_files

    def build_df(self, rows, index_type='datetime', continuous_data=1, categoric_data=1):
        """
            Build a random dataframe based on the input parameters
        """
        if index_type not in self.index_types:
            raise ValueError(f"Unsupported index type {index_type}")

        continuous_cols = []
        for _ in range(continuous_data):
            continuous_func = self._generate_sin_function()
            continuous_cols.append(np.vectorize(continuous_func)(pd.Series(np.arange(rows))))

        categoric_cols = []
        for _ in range(categoric_data):
            fake_word = lambda x: self.fake.word()
            categoric_cols.append(np.vectorize(fake_word)(pd.Series(np.arange(rows))))

        df_cols = {self.fake.word(): array for array in categoric_cols + continuous_cols}
        return pd.DataFrame(df_cols, index=self.index_types[index_type](rows))

    @staticmethod
    def _generate_sin_function():
        """
            Return a random sin function
        """
        multiplier = random.randint(1, 100)
        modifier = random.randint(1, 100)

        return lambda x: multiplier * math.sin(x) + modifier

    @staticmethod
    def _generate_datetime_index(rows):
        """
            Return a series of datetimes which can be used as an index
        """
        return pd.period_range("2016-01-01", freq='H', periods=rows)

    @staticmethod
    def _generate_counter_index(rows):
        """
            Return a series of numbers as an index
        """
        return np.arange(rows)

    @staticmethod
    def _generate_unique_categoric_index(rows):
        """
            Return a UUID that can be used as a row index
        """
        generate_id: Callable[[Any], UUID] = lambda x: uuid.uuid4()
        return np.vectorize(generate_id)(np.arange(rows))


if __name__ == "__main__":
    datagen = FakeDataGen()
    gen_rows = 10000
    start_time = time.time()
    datagen.build_df_to_file(gen_rows, path='mock_filesystem/new_test',
                             continuous_data=10, categoric_data=10, file_spread=2)
    print(f"Completed data generation in {time.time() - start_time} seconds")



from typing import Optional
import graphviz

from discovery.metadata import CatalogueMetadata


class SubGraphNode:
    children: []
    filesystem_full_path: str
    graph: graphviz.Digraph

    def __init__(self, filesystem_full_path, graph: Optional[graphviz.Graph] = None):
        self.children = []
        self.filesystem_full_path = filesystem_full_path
        self._set_graph(graph)

    def _set_graph(self, graph: Optional[graphviz.Graph]):
        if graph is None:
            self.graph = self._generate_graph()
        else:
            self.graph = graph

    def _generate_graph(self):
        subgraph = graphviz.Graph(name="cluster_" + self.filesystem_full_path)
        subgraph.attr(label=self.filesystem_full_path)
        subgraph.attr(color='grey')
        return subgraph

    def get_parent_path(self):
        split = self.filesystem_full_path.rsplit('/', 1)
        if 1 < len(split):
            return split[0]
        return ""

    def recursively_append_subgraphs(self):
        for child in self.children:
            child.recursively_append_subgraphs()
            self.graph.subgraph(child.graph)
        return


class Visualizer:
    engine: str

    root: SubGraphNode
    working_node: SubGraphNode
    observed_nodes: [SubGraphNode]

    def __init__(self, engine="fdp"):
        self.engine = engine
        self.observed_nodes = []
        self.root = SubGraphNode("", self._generate_root_graph())
        self.working_node = self.root
        self.observed_nodes.append(self.working_node)

    def _generate_root_graph(self):
        start_graph = graphviz.Graph(name="cluster_root", comment='File Visualization',
                                     node_attr={'shape': 'plaintext'})
        start_graph.attr(label="")
        start_graph.engine = self.engine
        return start_graph

    def draw(self, metadata: [CatalogueMetadata], filename: str):
        self._draw_metadata(metadata)
        # self.draw_relationships(metadata)
        self._finalize_result_graph(filename)

    def _draw_metadata(self, metadata: [CatalogueMetadata]):
        for datum in metadata:
            file_path = datum.data_manifest.get('path', 'undefined')
            self.working_node = self._determine_working_node(file_path)
            self._draw_metadatum(datum)

    # Don't use it, not implemented properly
    def _draw_relationships(self, metadata):
        for metadatum in metadata:
            for column in metadatum.columns:
                for relationship in column.relationships:
                    self.root.graph.edge(f"{metadatum.hash}:{column.name}",
                                         f"{relationship.target_file_hash}:{relationship.target_column_name}",
                                         str(relationship.certainty)
                                         )

    def _finalize_result_graph(self, output_filename):
        self.root.recursively_append_subgraphs()
        self.root.graph.view(output_filename)

    def _draw_metadatum(self, metadatum: CatalogueMetadata):
        columns = self._draw_table_columns(metadatum)
        filled_table = self._draw_filled_metadatum_table(metadatum, columns)
        self.working_node.graph.node(str(metadatum.item_id), filled_table)

    # TODO: make it more generic
    def _draw_table_columns(self, metadatum):
        col_rows: str = ""
        for column in metadatum.columns.values():
            col_rows += '<TR><TD>{}</TD> <TD>{}</TD> <TD>{}</TD> <TD>{}</TD> <TD>{}</TD> <TD>{}</TD> ' \
                .format(column.name, column.col_type,
                        str(round(column.continuity * 100, 2)) + '%',
                        getattr(column, "mean", "NA"),
                        getattr(column, "minimum", "NA"),
                        getattr(column, "maximum", "NA")
                        )
            # 'NA' if column.stationarity is None else ('Yes' if column.stationarity == 1 else 'No'))

            for relationship in column.relationships:
                col_rows += '<TD>' + relationship.target_column_name + ' ' + \
                            str(round(relationship.certainty, 2)) + '%</TD>'
            col_rows += '</TR>'
        return col_rows

    # TODO: find a more generic approach
    def _draw_filled_metadatum_table(self, metadatum, col_strings: str):
        file_path = metadatum.data_manifest.get('path', 'undefined')
        data_size = metadatum.data_manifest['data_size']
        filled_table = f'''<
            <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">
              <TR>
                <TD COLSPAN="8" BGCOLOR="darkgrey">{file_path}</TD>
              </TR>
              <TR>
                <TD COLSPAN="8" BGCOLOR="darkgrey">{data_size.get('rows', 0)} rows</TD>
              </TR>
              <TR>
                <TD COLSPAN="8" BGCOLOR="darkgrey">Tags: {', '.join(metadatum.tags)}</TD>
              </TR>
              <TR>
                <TD BGCOLOR="lightgray">Name</TD>
                <TD BGCOLOR="lightgray">Type</TD>
                <TD BGCOLOR="lightgray">Numeric</TD>
                <TD BGCOLOR="lightgray">Average</TD>
                <TD BGCOLOR="lightgray">Lowest</TD>
                <TD BGCOLOR="lightgray">Highest</TD>
                '''
        # <TD BGCOLOR="lightgray">Continuity</TD>
        # <TD BGCOLOR="lightgray">Stationarity</TD>

        for column in metadatum.columns.values():
            if column.relationships:
                # map the certainties to a key, multiple certainties will override each other, but we only care about one
                certainty_map = {relationship.certainty: relationship for relationship in column.relationships}
                highest_certainty = certainty_map[max(certainty_map)]
                filled_table += f"<TD BGCOLOR='lightgray'>Best column match in {highest_certainty.target_hash}</TD>"

        filled_table += f"</TR>{col_strings}</TABLE>>"

        return filled_table

    def _determine_working_node(self, filesystem_full_path: str):
        # determined_node = None

        already_visited_node = self._get_node_if_path_was_already_observed(filesystem_full_path)
        if already_visited_node is not None:
            determined_node = already_visited_node
        else:
            determined_node = SubGraphNode(filesystem_full_path)
            self._add_all_previously_unobserved_parents_to_observed_nodes(determined_node)
        return determined_node

    def _add_all_previously_unobserved_parents_to_observed_nodes(self, parameter_child_node: SubGraphNode):
        working_child_node = parameter_child_node
        while True:
            visited_parent = self._get_node_if_path_was_already_observed(working_child_node.filesystem_full_path)
            if visited_parent is None:
                new_parent = SubGraphNode(working_child_node.get_parent_path())
                new_parent.children.append(working_child_node)
                # setup next loop
                working_child_node = new_parent
            else:
                visited_parent.children.append(working_child_node)
                break

    def _get_node_if_path_was_already_observed(self, filesystem_full_path: str):
        return next(iter([node for node in self.observed_nodes if node.filesystem_full_path == filesystem_full_path]),
                    None)
